import cv2
import numpy as np
import math
from skimage import morphology, measure
from scipy import spatial
from sklearn.cluster import MeanShift
from sklearn.neighbors import NearestNeighbors
import copy


# class ROI: Aims to record basic ROI properties which can be used in all following sequences analysis
# Input: the file address of marked ROI
# Object element:
# 1. Im_OOI: 2D image of ROI, np.binary
# 2. skeleton_point: skeleton points of ROI, point format: [y, x]
# 3. skeleton_distance: path distance computed by skeleton points
# 4. min_y, min_x, max_y, max_x: the max and min value of x and y range
class ROI:
    def __init__(self, marked_image_root):
    ## read the manually marked ROI
        Im_marked = cv2.imread(marked_image_root)
    ## determine the ROI and extract ROI properties
        self.Im_OOI = np.logical_and(np.equal(Im_marked[:, :, 0], 0), np.equal(Im_marked[:, :, 1], 255))
        self.Im_OOI = np.logical_and(np.equal(Im_marked[:, :, 2], 0), self.Im_OOI)
        self.min_y = np.min(np.nonzero(self.Im_OOI)[0])
        self.min_x = np.min(np.nonzero(self.Im_OOI)[1])
        self.max_y = np.max(np.nonzero(self.Im_OOI)[0])
        self.max_x = np.max(np.nonzero(self.Im_OOI)[1])
    ## skeletonize the ROI and extract skeleton properties
        Im_skeleton = morphology.skeletonize(self.Im_OOI)
        Im_skeleton_labels = measure.label(Im_skeleton, connectivity=2)
        Im_skeleton_prop = measure.regionprops(Im_skeleton_labels)
        self.skeleton_point = Im_skeleton_prop[0].coords
        self.skeleton_distance = spatial.distance.pdist(self.skeleton_point)
        self.skeleton_distance = spatial.distance.squareform(self.skeleton_distance)


# class MC_particle: Single MC particle, with properties, position in 2D, velocity, weight
# initialization: get the init position of MC particle
# eff_check: check it is in OOI
# intensity_move: move particle into the brighter spots in the neighbour
# velocity_move: MC particle movement based on the transition function
# weight_compute: weight computation based on the grayscale value
class MC_particle:
    def __init__(self, expected_intensity=100, intensity_sigma=30, min_intensity=50):
        self.particle_single_x = -1
        self.particle_single_y = -1
        self.weight = -1
        self.velocity = 0
        self.expected_intensity = expected_intensity
        self.intensity_sigma = intensity_sigma
        self.min_intensity = min_intensity

    def initialization(self, OOI):
        while(1):
            particle_single_x = np.random.randint(low=OOI.min_x, high=OOI.max_x, size=1)[0]
            particle_single_y = np.random.randint(low=OOI.min_y, high=OOI.max_y, size=1)[0]
            if self.eff_check(OOI, particle_single_x, particle_single_y):
                self.particle_single_x = particle_single_x
                self.particle_single_y = particle_single_y
                break

    def eff_check(self, OOI, x, y):
        if OOI.Im_OOI[y, x] == 1:
            return True
        else:
            return False

    def intensity_move(self, Frame, search_step):
        max_grayscale = Frame.Im_postprocessed[self.particle_single_y, self.particle_single_x]
        max_y_offest = 0
        max_x_offset = 0
        for i in range(-search_step, search_step + 1, 1):
            for j in range(-search_step, search_step + 1, 1):
                if self.particle_single_y + i < Frame.Im_postprocessed.shape[0] and \
                        self.particle_single_x + j < Frame.Im_postprocessed.shape[1]:
                    if Frame.Im_postprocessed[self.particle_single_y + i, self.particle_single_x + j] > max_grayscale:
                        max_grayscale = Frame.Im_postprocessed[self.particle_single_y + i, self.particle_single_x + j]
                        max_y_offest = i
                        max_x_offset = j
        self.particle_single_y = self.particle_single_y + max_y_offest
        self.particle_single_x = self.particle_single_x + max_x_offset

    def velocity_move(self, OOI, min_velocity, max_velocity, vessel_dir, velocity_error, Frame, search_step):
        neigh = NearestNeighbors(n_neighbors=1)
        neigh.fit(OOI.skeleton_point)
        particle_single_velocity = np.random.uniform(low=min_velocity, high=max_velocity, size=1)[0]

        if vessel_dir == 0:
            particle_single_velocity = -particle_single_velocity
            particle_min_velocity = particle_single_velocity + velocity_error
            particle_max_velocity = particle_single_velocity - velocity_error
        else:
            particle_single_velocity = particle_single_velocity
            particle_min_velocity = particle_single_velocity - velocity_error
            particle_max_velocity = particle_single_velocity + velocity_error

        for _ in range(10000):
            particle_single_x = np.random.randint(low=OOI.min_x, high=OOI.max_x, size=1)[0]
            particle_single_y = np.random.randint(low=OOI.min_y, high=OOI.max_y, size=1)[0]
            if vessel_dir == 0:
                if particle_single_x > self.particle_single_x:
                    continue
            if vessel_dir == 1:
                if particle_single_x < self.particle_single_x:
                    continue
            if not self.eff_check(OOI, particle_single_x, particle_single_y):
                continue
            else:
                potential_velocity = math.pow((particle_single_x - self.particle_single_x), 2) + math.pow((particle_single_y - self.particle_single_y), 2)
                if potential_velocity < math.pow(particle_max_velocity, 2) and potential_velocity > math.pow(particle_min_velocity, 2):
                    if not self.eff_check(OOI, particle_single_x, particle_single_y):
                        continue
                    sample = np.asarray([[self.particle_single_y, self.particle_single_x]])
                    old_idx = neigh.kneighbors(sample, return_distance=True).__getitem__(1)[0, 0]
                    MC_particle_temp = MC_particle()
                    MC_particle_temp.particle_single_x = particle_single_x
                    MC_particle_temp.particle_single_y = particle_single_y
                    MC_particle_temp.intensity_move(Frame, search_step)
                    particle_single_x = MC_particle_temp.particle_single_x
                    particle_single_y = MC_particle_temp.particle_single_y
                    sample = np.asarray([[particle_single_y, particle_single_x]])
                    new_idx = neigh.kneighbors(sample, return_distance=True).__getitem__(1)[0, 0]
                    if vessel_dir == 0:
                        velocity = -OOI.skeleton_distance[new_idx, old_idx]
                    else:
                        velocity = OOI.skeleton_distance[new_idx, old_idx]
                    MC_particle_temp = MC_particle()
                    MC_particle_temp.particle_single_x = particle_single_x
                    MC_particle_temp.particle_single_y = particle_single_y
                    MC_particle_temp.velocity = velocity
                    return 1, MC_particle_temp
        return 0, 0

    def weight_compute(self, Frame):
        gray_value = Frame.Im_postprocessed[self.particle_single_y, self.particle_single_x]
        if gray_value < self.min_intensity:
            self.weight = 0
        else:
            self.weight = math.exp(
                -min((gray_value - self.expected_intensity), 0) * min((gray_value - self.expected_intensity), 0) /
                (2 * self.intensity_sigma * self.intensity_sigma))


# class MC_particle_cluster: Cluster of MC_particles, represent single RBC.
# add_particle: add MC_particle into the cluster
# intensity_move: intensity move all MC_particles in the cluster
# velocity_move: velocity move all MC_particles in the cluster
# weight_compute: compute weights of all MC_particles in the cluster
# weight_sum: compute the sum of MC_particle weights in the cluster
# weight_compute_norm: normalize the MC_particle weights
# resampling: resampling MC_particle in the cluster
# clustering: meanshift clustering

class MC_particle_cluster:
    def __init__(self, cluster_idx):
        self.cluster_idx = cluster_idx
        self.particle_list = []

    def add_particle(self, MC_particle):
        self.particle_list.append(MC_particle)

    def intensity_move(self, Frame, search_step):
        for i in range(self.particle_list.__len__()):
            self.particle_list[i].intensity_move(Frame, search_step)

    def velocity_move(self, OOI, min_velocity, max_velocity, vessel_dir, velocity_error, Frame, search_step):
        output_particle_list = []
        for i in range(self.particle_list.__len__()):
            eff_flag, particle_temp = self.particle_list[i].velocity_move(OOI, min_velocity, max_velocity,
                                                           vessel_dir, velocity_error, Frame, search_step)
            if eff_flag:
                output_particle_list.append(particle_temp)

        self.particle_list = output_particle_list
        if self.particle_list.__len__() <= 0:
            return 0
        else:
            return 1

    def weight_compute(self, Frame):
        max_weight = -1
        for i in range(self.particle_list.__len__()):
            self.particle_list[i].weight_compute(Frame)
            if self.particle_list[i].weight > max_weight:
                max_weight = self.particle_list[i].weight
        if max_weight <= 0:
            return False
        else:
            return True

    def weight_sum(self):
        sum = 0
        for i in range(self.particle_list.__len__()):
            sum = sum + self.particle_list[i].weight
        return sum

    def weight_compute_norm(self, resample_flag = False):
        if not resample_flag:
            particle_weights_sum = self.weight_sum()
            updated = []
            for i in range(self.particle_list.__len__()):
                updated.append(self.particle_list[i].weight / particle_weights_sum)

            for i in range(self.particle_list.__len__()):
                self.particle_list[i].weight = updated[i]
        else:
            for i in range(self.particle_list.__len__()):
                self.particle_list[i].weight = 1.0 / self.particle_list.__len__()

    def compute_pos_average(self):
        pos_avg_x = 0
        pos_avg_y = 0
        for i in range(self.particle_list.__len__()):
            pos_avg_x = self.particle_list[i].particle_single_x + pos_avg_x
            pos_avg_y = self.particle_list[i].particle_single_y + pos_avg_y
        pos_avg_x = pos_avg_x / self.particle_list.__len__()
        pos_avg_y = pos_avg_y / self.particle_list.__len__()
        return pos_avg_x, pos_avg_y

    def compute_velocity_average(self):
        avg_v = 0
        for i in range(self.particle_list.__len__()):
            avg_v = self.particle_list[i].velocity + avg_v
        avg_v = avg_v / self.particle_list.__len__()
        return avg_v

    def resampling(self, particle_number):
        particle_list_temp = []
        particle_weight_input = []
        for i in range(self.particle_list.__len__()):
            particle_weight_input.append(self.particle_list[i].weight)
        particle_weight_input = np.asarray(particle_weight_input)

        particles_cumulative_weights = np.zeros((particle_weight_input.shape[0],), dtype=np.float32)
        for i in range(particle_weight_input.shape[0]):
            particles_cumulative_weights[i] = np.sum(particle_weight_input[0:i + 1])

        for j in range(particle_number):
            random_num = np.random.rand(1)
            particles_index = np.greater_equal(particles_cumulative_weights, random_num[0])
            particles_index = np.nonzero(particles_index)
            particles_index = np.asarray(particles_index)
            particle_resample_single = self.particle_list[particles_index[0, 0]]
            particle_list_temp.append(particle_resample_single)
        self.particle_list = particle_list_temp

    def clustering(self, Frame, MeanShift_bandwidth, min_particles_in_cluster, min_component_weights):
        particle_pos_flatten = []
        total_number_of_particles = self.particle_list.__len__()
        for i in range(self.particle_list.__len__()):
            particle_pos_flatten.append((self.particle_list[i].particle_single_y, self.particle_list[i].particle_single_x))
        particle_pos_flatten = np.asarray(particle_pos_flatten)
        clustering = MeanShift(bandwidth=MeanShift_bandwidth).fit(particle_pos_flatten)

        for i in range(np.max(clustering.labels_) + 1):
            particle_num_in_cluster = np.count_nonzero(np.equal(clustering.labels_, i))
            if particle_num_in_cluster < min_particles_in_cluster:
                for ii in range(clustering.labels_.shape[0]):
                    if clustering.labels_[ii] == i:
                        clustering.labels_[ii] = -1

        component_weights = []
        particle_clusters_output_temp = []
        for i in range(np.max(clustering.labels_) + 1):
            particle_cluster_temp = MC_particle_cluster(-1)
            for j in range(clustering.labels_.shape[0]):
                if clustering.labels_[j] == i:
                    particle_cluster_temp.add_particle(copy.deepcopy(self.particle_list[j]))
            particle_cluster_temp.weight_compute(Frame)
            component_weights.append(particle_cluster_temp.weight_sum())
            particle_clusters_output_temp.append(particle_cluster_temp)

        component_weights = np.asarray(component_weights)
        component_weights = component_weights / np.sum(component_weights)
        particle_clusters_output = []
        component_weights_output = []
        for i in range(np.max(clustering.labels_) + 1):
            if component_weights[i] >= min_component_weights:
                component_weights_output.append(component_weights[i])
                particle_clusters_output.append(particle_clusters_output_temp[i])

        component_weights_output = np.asarray(component_weights_output)
        component_num = np.uint16(np.ceil(component_weights_output * total_number_of_particles / np.sum(component_weights_output)))
        for i in range(0, particle_clusters_output.__len__()):
            particle_clusters_output[i].weight_compute_norm(resample_flag=False)
            particle_clusters_output[i].resampling(component_num[i])
            particle_clusters_output[i].weight_compute_norm(resample_flag=True)
        return particle_clusters_output




# class MC_frame: Aims to record MC particle cluster information in each frame
# particle_init: Init MC particles in the frame
# compute_maximum_cluster_idx: compute a the largest cluster idx in the current frame
# add_cluster_from_clusters: add new MC particle cluster from a list of running particles

class MC_frame:
    def __init__(self, frame_idx):
        self.frame_idx = frame_idx
        self.particle_cluster_list = []

    def particle_init(self, particles_total, OOI):
        # All the MC particles are considered as one RBC cluster when initialization
        running_particles_clusters = MC_particle_cluster(-1)
        for _ in range(particles_total):
            single_particle = MC_particle()
            single_particle.initialization(OOI)
            running_particles_clusters.add_particle(single_particle)
        return running_particles_clusters

    def compute_maximum_cluster_idx(self):
        max_cluster_idx = 0
        for i in range(self.particle_cluster_list.__len__()):
            if self.particle_cluster_list[i].cluster_idx > max_cluster_idx:
                max_cluster_idx = self.particle_cluster_list[i].cluster_idx
        return max_cluster_idx

    def add_cluster_from_clusters(self, particle_clusters, former_max, cts_flag):
        current_class_centers = []
        current_clusters = copy.deepcopy(self.particle_cluster_list)

        for i in range(self.particle_cluster_list.__len__()):
            cluster_center_x, cluster_center_y = self.particle_cluster_list[i].compute_pos_average()
            current_class_centers.append((cluster_center_y, cluster_center_x))


        current_class_centers = np.asarray(current_class_centers)
        for i in range(particle_clusters.__len__()):
            new_cluster_flag = 1
            new_center_x, new_center_y = particle_clusters[i].compute_pos_average()

            for j in range(self.particle_cluster_list.__len__()):
                center_distance = (new_center_x - current_class_centers[j, 1])*(new_center_x - current_class_centers[j, 1])
                center_distance = center_distance + (new_center_y - current_class_centers[j, 0])*(new_center_y - current_class_centers[j, 0])
                center_distance = math.sqrt(center_distance)
                if center_distance < (3*1.414) or self.particle_cluster_list[j].cluster_idx == particle_clusters[i].cluster_idx:
                    new_cluster_flag = 0
                    particle_clusters[i] = copy.deepcopy(self.particle_cluster_list[j])
                    break
            if new_cluster_flag == 1:
                if not cts_flag:
                    particle_clusters[i].cluster_idx = former_max + 1
                    former_max = former_max + 1
                current_clusters.append(particle_clusters[i])
        self.particle_cluster_list = copy.deepcopy(current_clusters)
        return particle_clusters



# class Im_frame: Aims to read raw image, preprocess, and draw
# Initialization: read frame, minus temporal average
# Im_postprocessing: Get frame variance map
# draw_particles_all: draw all MC particles
# draw_particle_clusters: draw center points of moving MC particle clusters
# draw_particle_clusters_all: draw all MC particles with cluster infomation.

class Im_frame:
    def initialization(self, frame_idx, raw_image_root, temporal_avg):
        Im_name = raw_image_root + str(frame_idx) + '.tif'
        self.Im_raw = cv2.imread(Im_name, cv2.IMREAD_GRAYSCALE)
        self.Im_preprocessed = np.subtract(np.double(self.Im_raw), np.double(temporal_avg))
        self.Im_preprocessed = np.uint8(np.minimum(np.maximum(self.Im_preprocessed, 0), 255))
        self.Im_postprocessed = self.Im_preprocessed


    def Im_postprocessing(self, former_frame):
        self.Im_postprocessed = np.subtract(np.double(self.Im_preprocessed), np.double(former_frame.Im_preprocessed))
        self.Im_postprocessed = np.uint8(np.minimum(np.maximum(self.Im_postprocessed, 0), 255))

    def draw_particles_all(self, particle_cluster, window_name, color=(0, 0, 255)):
        im_vis = np.zeros((self.Im_raw.shape[0], self.Im_raw.shape[1], 3), dtype=np.uint8)
        im_vis[:, :, 0] = self.Im_raw
        im_vis[:, :, 1] = self.Im_raw
        im_vis[:, :, 2] = self.Im_raw
        for j in range(particle_cluster.particle_list.__len__()):
            cv2.circle(im_vis, center=(particle_cluster.particle_list[j].particle_single_x,
                                       particle_cluster.particle_list[j].particle_single_y),
                       radius=1, color=color, thickness=1)
        cv2.imshow(window_name, im_vis)
        cv2.waitKey()

    def draw_particle_clusters(self, particle_clusters, window_name, radius=1, save_flag = 0):
        im_vis = np.zeros((self.Im_raw.shape[0], self.Im_raw.shape[1], 3), dtype=np.uint8)
        im_vis[:, :, 0] = self.Im_raw
        im_vis[:, :, 1] = self.Im_raw
        im_vis[:, :, 2] = self.Im_raw
        color_map = []
        # red
        color_map.append((0, 0, 255))
        # green
        color_map.append((0, 255, 0))
        # blue
        color_map.append((255, 0, 0))
        # yellow
        color_map.append((0, 255, 255))
        # purple
        color_map.append((255, 0, 255))
        # cyan
        color_map.append((255, 255, 0))

        color_map.append((0, 0, 128))
        color_map.append((0, 128, 0))
        color_map.append((128, 0, 0))
        color_map.append((0, 128, 128))
        color_map.append((128, 0, 128))
        color_map.append((128, 128, 0))

        effective_idx = []
        for i in range(particle_clusters.__len__()):
            idx = (particle_clusters[i].cluster_idx - 1)%12
            color = color_map[idx]
            pos_avg_x, pos_avg_y = particle_clusters[i].compute_pos_average()
            if abs(particle_clusters[i].compute_velocity_average())>0:
                effective_idx.append(particle_clusters[i].cluster_idx)
                cv2.circle(im_vis, center=(np.uint32(pos_avg_x), np.uint32(pos_avg_y)),
                           radius=radius, color=color, thickness=1)
        if not save_flag:
            cv2.imshow(window_name, im_vis)
            cv2.waitKey()
        else:
            cv2.imwrite('../image/results/' + window_name + '.tif', im_vis)
        return effective_idx, color_map

    def draw_particle_clusters_all(self, particle_clusters, window_name, radius=1):
        im_vis = np.zeros((self.Im_raw.shape[0], self.Im_raw.shape[1], 3), dtype=np.uint8)
        im_vis[:, :, 0] = self.Im_raw
        im_vis[:, :, 1] = self.Im_raw
        im_vis[:, :, 2] = self.Im_raw
        color_map = []
        # red
        color_map.append((0, 0, 255))
        # green
        color_map.append((0, 255, 0))
        # blue
        color_map.append((255, 0, 0))
        # yellow
        color_map.append((0, 255, 255))
        # purple
        color_map.append((255, 0, 255))
        # cyan
        color_map.append((255, 255, 0))

        color_map.append((0, 0, 128))
        color_map.append((0, 128, 0))
        color_map.append((128, 0, 0))
        color_map.append((0, 128, 128))
        color_map.append((128, 0, 128))
        color_map.append((128, 128, 0))

        for i in range(particle_clusters.__len__()):
            if i < 12:
                color = color_map[i]
            else:
                color = color_map[11]
            for j in range(particle_clusters[i].particle_list.__len__()):
                cv2.circle(im_vis, center=(particle_clusters[i].particle_list[j].particle_single_x,
                                           particle_clusters[i].particle_list[j].particle_single_y),
                           radius=1, color=color, thickness=1)
        cv2.imshow(window_name, im_vis)
        cv2.waitKey()

# discard low weight clusters from cluster list
def clusters_discard(input_clusters, min_component_weights, particles_total):
    sum_1 = []
    for i in range(input_clusters.__len__()):
        sum_1.append(input_clusters[i].weight_sum())
    sum_1 = np.asarray(sum_1)
    sum_1 = sum_1 / np.sum(sum_1)

    output_clusters = []
    sum_2 = []
    for i in range(input_clusters.__len__()):
        if sum_1[i] >= min_component_weights:
            output_clusters.append(input_clusters[i])
            sum_2.append(sum_1[i])

    sum_2 = np.asarray(sum_2)
    sum_2 = sum_2 * particles_total / np.sum(sum_2)

    for i in range(output_clusters.__len__()):
        output_clusters[i].weight_compute_norm(resample_flag=False)
        output_clusters[i].resampling(np.uint32(np.ceil(sum_2[i])))

    return output_clusters